theory Basics
  imports Main (*imports Isabelle/HOL's standard library*)
begin

section \<open>Text Structure and Navigation\<close>

text \<open>
  \<bullet> SideKick provides a tree view of various file formats

  \<bullet> navigation works via Back and Forward buttons, like in a web browser

  \<bullet> text folds can be manipulated systematically (e.g. to collapse subproofs)

  \<bullet> CONTROL-hover (on Linux/Windows) or COMMAND-hover (on Mac OS X)
    reveals additional information, notably hyper-links, e.g.,
    \<bullet> @{url "http://isabelle.in.tum.de"}
    \<bullet> @{dir "~~/src/HOL/Library"}
    \<bullet> @{term "x + 1"}
\<close>


section \<open>Documentation\<close>

text \<open>
  notable manuals (see also the "Documentation panel"):

  \<bullet> jEdit manual (F1),

  \<bullet> @{doc "prog-prove"}, a concrete and example based introduction to Isabelle/HOL

  \<bullet> @{doc "jedit"}, Isabelle/jEdit manual

  \<bullet> @{doc "isar-ref"}, the main reference to Isabelle's
    Isar ("Intelligible Semi-Automated Reasoning") language (e.g., the content of this file)

  \<bullet> @{doc "main"}: a PDF version of Isabelle/HOL's "standard library"
\<close>


section \<open>Theory Files\<close>

text \<open>
  \<bullet> structure:
    \<open>
      theory Thy_Name -- \<open>in file "Thy_Name.thy"\<close>
        imports T\<^sub>1 ... T\<^sub>n -- \<open>imported theories; basis for Isabelle/HOL: Main\<close>
      begin
        ... -- \<open>theory content\<close>
      end
    \<close>

  \<bullet> content belongs to "outer syntax" (@{doc "isar-ref"} Chapter 3), i.e.,
    syntax of Isabelle/Isar ("Intelligible Semi-Automated Reasoning") theory sources,
    consisting of "commands", "methods", and "attributes"
    \<bullet> commands for specifying definitions, functions, inductive sets, ... (later)
    \<bullet> commands for stating propositions, initiating proofs, ... (see below)
    \<bullet> diagnostic commands (see below)
    \<bullet> methods for proving (see below and later)
    \<bullet> attributes for operations on "facts" (later)
    ...

  \<bullet> "inner syntax" refers to types and terms of the logic and is embedded
    into outer syntax by double-quotes ('"', '"') or cartouches, if not atomic

  \<bullet> example: @{file "~~/src/HOL/ex/Seq.thy"}
\<close>

term -- \<open>outer syntax: diagnostic command to show terms\<close>
  "\<lambda>x. x" -- \<open>inner syntax: a \<lambda>-term denoting the identity function\<close>
term \<open>\<lambda>x. x\<close> -- \<open>the same using cartouches\<close>


section \<open>Isabelle Symbols\<close>

text \<open>
  \<bullet> Isabelle supports infinitely many mathematical symbols: \<foobar>, \<baz>

  \<bullet> finitely many symbols are rendered via Unicode: e.g. "bar" as "\<bar>"

  \<bullet> some notable symbols with abbreviations for input (see also Symbols panel):

    symbol    name with backslash    abbreviation
    \<lambda>         \lambda                %
    \<Rightarrow>        \Rightarrow            =>
    \<Longrightarrow>       \Longrightarrow        ==>
    \<And>         \And                   !!
    \<equiv>         \equiv                 ==
    \<forall>         \forall                !
    \<exists>         \exists                ?
    \<longrightarrow>       \longrightarrow        -->
    \<and>         \and                   &
    \<or>         \or                    |
    \<not>         \not                   ~
    \<noteq>         \noteq                 ~=
    \<in>         \in                    :
    \<notin>         \notin                 ~:
    \<circ>         \circ
    \<cdot>          \cdot
    \<langle>          \langle               <<
    \<rangle>          \rangle               >>

  \<bullet> binder syntax: \<open>binder_symbol x\<^sub>1 ... x\<^sub>n. P x\<^sub>1 ... x\<^sub>n\<close>
\<close>

subsection \<open>Exercise\<close>

text \<open>
  \<bullet> write some logical propositions using Isabelle symbols
\<close>

-- \<open>diagnostic command for showing propositions\<close>
prop "A"


section \<open>Logical Language: Isabelle/HOL -- "The Object Logic"\<close>

text \<open>
  main kinds of formal entities:

  \<bullet> types: simple

  \<bullet> terms: \<lambda>-calculus with \<alpha>\<beta>\<eta>-equality

  \<bullet> propositions: terms of type bool

  \<bullet> theorems are proven propositions

  \<bullet> facts are lists of theorems
\<close>

-- \<open>diagnostic command to show types\<close>
typ unit
typ nat
typ bool
typ "'a \<times> 'b"
typ "'a \<Rightarrow> 'b"
typ "'a list"
typ "'a option"

-- \<open>diagnostic command to show terms\<close>
term "x :: nat"
term Suc
term "\<lambda>y. x + Suc y"
term "[1, 2, 3]"
term "Some ()"

-- \<open>diagnostic command to show propositions\<close>
prop "x = y"
prop "A \<longrightarrow> B"
prop "\<forall>x. A x \<longrightarrow> B x"

-- \<open>diagnostic command to show facts\<close>
thm refl sym trans
thm conjI conjunct1 conjunct2
thm impI mp
thm allI spec exI exE


section \<open>Constants and Variables: black, blue, green, what?\<close>

text \<open>
  \<bullet> constants are printed black as in @{term "True"}

  \<bullet> bound variables are printed green as in @{term "\<lambda>x. x"}

  \<bullet> free variables are printed blue as in @{term "f x"}

  \<bullet> both of the above are "fixed" (no substitution allowed)

  \<bullet> schematic variables are prefixed by a question mark and allow for arbitrary substitutions

  \<bullet> all free variables of a proven statement are replaced by schematic ones
\<close>


section \<open>Rule Statements: Isabelle/Pure -- "The Meta Logic"\<close>

text \<open>
  \<bullet> Isabelle/Pure (@{theory Pure}) provides minimal logic
    connectives, to express Natural Deduction rule statements
\<close>

-- \<open>Pure implication: logical entailment\<close>
prop "A \<Longrightarrow> B"
-- \<open>Pure all-quantification: local "eigen-variables"\<close>
prop "\<And>x::'a. B x"

 -- \<open>rule with premises \<open>A\<^sub>1\<close>, \<open>A\<^sub>2\<close>, \<open>A\<^sub>3\<close> and conclusion \<open>C\<close>\<close>
prop "A\<^sub>1 \<Longrightarrow> A\<^sub>2 \<Longrightarrow> A\<^sub>3 \<Longrightarrow> C"
 -- \<open>nested rule: assumption demands subproof with local context\<close>
prop "(\<And>x. A x \<Longrightarrow> B x) \<Longrightarrow> C"


text \<open>
  \<bullet> Isabelle/Isar provides separate notation for structured rule statements
\<close>

thm impI mp
thm allI spec exI exE

print_statement impI mp
print_statement allI spec exI exE


section \<open>Proofs\<close>

text \<open>
  \<bullet> atomic proofs:
    \<bullet> skip proofs with @{command sorry}
    \<bullet> single step proofs @{command by} (as in "by method")
    \<bullet> double step proofs (as in
      "by (initial_method) (terminal_method)"),
      e.g., "by (induct n) (auto)"

  \<bullet> goal refinement / subproofs:
    \<bullet> proof (initial_method)
       ...
      qed (optional_terminal_method)
\<close>

notepad -- \<open>a scratchpad for proof exploration without specific goal\<close>
begin
  text \<open>
    \<bullet> "raw proof blocks" are enclosed by braces ('{', '}')
    \<bullet> when closed, they result in rule statements
  \<close>
  { -- \<open>"meta" implications\<close>
    assume "A"
    have "B" sorry
  } -- \<open>result: \<open>A \<Longrightarrow> B\<close>\<close>
  have "A \<Longrightarrow> B" by fact -- \<open>reprove facts that are already known (locally)\<close>

  { -- \<open>reasoning about "arbitrary but fixed" values\<close>
    fix x :: 'a
    have "P x" sorry
  } -- \<open>result: (almost) equivalent to \<open>\<And>x. P x\<close>\<close>
  have "\<And>x. P x" by fact
end

text \<open>
  proof state consists of a number of "subgoals" of the form

    \<open>A\<^sub>1 \<Longrightarrow> ... \<Longrightarrow> A\<^sub>n \<Longrightarrow> C\<close>

  where \<open>A\<^sub>1\<close> to \<open>A\<^sub>n\<close> are called "assumptions"
  and \<open>C\<close> is called "conclusion"
\<close>


section \<open>Proof Methods\<close>

text \<open>
  • @{method rule}: generic Natural Deduction (with
    HO unification)

  • @{method cases}: case analysis

  • @{method induct} and @{method coinduct}: induction
    and coinduction

  • @{method simp}: equational reasoning by the
    Simplifier (HO rewriting)

  • @{method fast} and @{method blast}:
    classical reasoning (tableau)

  • @{method auto} and @{method force}:
    combined simplification and classical reasoning

  • @{method arith}, @{method presburger}:
    specific theories

  • ...
\<close>


section \<open>Application of Rules: Simplification, Introduction, Elimination, Destruction\<close>

lemma "(\<exists>x. \<forall>y. P x y) \<longrightarrow> (\<forall>y. \<exists>x. P x y)"
text \<open>
  \<bullet> as intro(duction) rule
  \<bullet> \<open>rule\<close> without "chained" facts
  \<bullet> conclusion of subgoal has to match conclusion of rule
  \<bullet> assumptions of rule turn into new subgoals
\<close>
thm impI
proof (rule impI)
  assume "\<exists>x. \<forall>y. P x y"
  text \<open>
    \<bullet> as elim(ination) rule
    \<bullet> \<open>rule\<close> with "chained" facts
    \<bullet> chained facts have to match major premise(s) of rule
    \<bullet> conclusion of subgoal always matches conclusion of rule
    \<bullet> remaining premises turn into new subgoals
  \<close>
  thm exE
  then -- \<open>chain previous fact into proposition\<close>
  show "\<forall>y. \<exists>x. P x y"
  proof (rule exE)
    fix x assume "\<forall>y. P x y"
    show "\<forall>y. \<exists>x. P x y"
    -- \<open>implicit initial rule application: but which rule and how?\<close>
    proof
      fix y
      text \<open>
        \<bullet> as dest(ruction) rule
        \<bullet> \<open>rule\<close> with chained facts
        \<bullet> chained facts have to match major premises of rule and
        \<bullet> conclusion of subgoal has to match conclusion of rule
      \<close>
      from \<open>\<forall>y. P x y\<close> -- \<open>literal fact\<close>
      have "P x y"
         -- \<open>remember: by (method) == proof (method) qed\<close>
        by (rule spec)
      then show "\<exists>x. P x y" .. -- \<open>implicit final rule application\<close>
    qed -- \<open>prove current subgoal\<close>
  qed
qed

lemma "hd (tl [1,2,3]) = 2"
  text \<open>
    \<bullet> as simp(lification) rule
    \<bullet> \<open>simp\<close> or \<open>unfolding\<close>
    \<bullet> instances of left-hand side of rule are exhaustively
      replaced by instances of right-hand side
    \<bullet> simp rules may be guarded by conditions (which are
      also discharged by simplification)
  \<close>
  thm list.sel
  -- \<open>exhaustively "rewrite" with rule from left to right\<close>
  unfolding list.sel(3)
  by (simp add: list.sel(1))

end
