theory Loop_Checker_Demo
  imports
    Trs
    Matching
begin

section \<open>Implementation\<close>

text \<open>
  The goal is to write a function

    @{term "check_loop :: ('f, 'v) trs \<Rightarrow> _ \<Rightarrow> bool"}

  which checks that a given TRS admits a loop, i.e.,
  a derivation of the form \<open>t \<rightarrow>\<^sup>+ C[t\<sigma>]\<close>
\<close>


-- \<open>Witness/certificate for single rewrite steps: rule + context\<close>
type_synonym ('f, 'v) rstep_wit = "('f, 'v) rule \<times> ('f, 'v) ctxt"

text \<open>
  Given \<open>t\<close> and \<open>C\<close> such that \<open>t = C[u]\<close> for some \<open>u\<close>,
  obtain subterm \<open>u\<close>.

  Used to obtain redex corresponding to rule application.
\<close>
fun get_redex :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
  where
    "get_redex (More f ss C _) (Fun g ts) = (get_redex C (ts ! length ss))"
  | "get_redex C t = t"

text \<open>Check single rewrite step\<close>
fun check_rstep_wit ::
  "('f, 'v) trs \<Rightarrow> ('f, 'v) rstep_wit \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    "check_rstep_wit R ((l, r), C) s t \<longleftrightarrow>
      (case match l (get_redex C s) of
        None \<Rightarrow> False
      | Some \<sigma> \<Rightarrow>
        s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<and> (l, r) \<in> R)"

fun check_derivation ::
  "('f, 'v) trs \<Rightarrow>
   ('f, 'v) term \<Rightarrow> (* source term *)
   (('f, 'v) rstep_wit \<times> ('f, 'v) term) list \<Rightarrow> (* list of steps and intermediate terms *)
   ('f, 'v) term option" (* on success: last term of derivation *)
  where
    "check_derivation R s [] = Some s"
  | "check_derivation R s ((wit, t) # seq) =
      (if check_rstep_wit R wit s t
        then check_derivation R t seq
        else None)"

-- \<open>substitutions as lists of bindings\<close>
type_synonym ('f, 'v) subst_impl = "('v \<times> ('f, 'v) term) list"

-- \<open>loop: \<open>t \<rightarrow> \<dots> \<rightarrow> C[t\<sigma>]\<close>\<close>
datatype ('f, 'v) loop_proof =
  Loop_Proof
    "('f, 'v) term" -- \<open>t\<close>
    "(('f, 'v) rstep_wit \<times> ('f, 'v) term) list" -- \<open>derivation\<close>
    "('f, 'v) ctxt" -- \<open>C\<close>
    "('f, 'v) subst_impl" -- \<open>\<sigma>\<close>

fun check_loop :: "('f, 'v) trs \<Rightarrow> ('f, 'v) loop_proof \<Rightarrow> bool"
  where
    "check_loop R (Loop_Proof t seq C s) = (
      let \<sigma> = som (map_of s) in
      seq \<noteq> [] \<and>
      (case check_derivation R t seq of
        None \<Rightarrow> False
      | Some u \<Rightarrow> u = C\<langle>t \<cdot> \<sigma>\<rangle>))"

definition main_checker ::
  "(string, string) trs \<Rightarrow> (string, string) loop_proof \<Rightarrow> string"
  where
    "main_checker R prf = (if check_loop R prf
      then ''certified nontermination''
      else ''rejected'')"

end
