theory Matching_Demo
  imports Term
begin

section \<open>Matching\<close>

text \<open>
  Given a "pattern" term \<open>s\<close> and another term \<open>t\<close>, matching
  is the problem of finding a substitution such that
  \<open>s \<cdot> \<sigma> = t\<close>.
\<close>

text \<open>
  Partial functions or "maps" from \<open>'a\<close> to \<open>'b\<close> are represented
  by @{typ "'a \<Rightarrow> 'b option"}. Some useful constants and facts
  regarding maps reside in @{theory Map}:

  \<bullet> @{const fun_upd} for updating an existing map with special
    syntax @{term "m(k \<mapsto> v)"} for adding a binding for
    value \<open>v\<close> with key \<open>k\<close>

  \<bullet> @{const empty} for the empty map

  \<bullet> @{const dom} for the "domain" of a map, i.e., the set of
    keys for which there are bindings

  \<bullet> @{const map_le} with syntax @{term "m\<^sub>1 \<subseteq>\<^sub>m m\<^sub>2"} ensures
    that \<open>m\<^sub>2\<close> is an extension of \<open>m\<^sub>1\<close>
\<close>

subsection \<open>Mutual Recursive Functions\<close>

fun even :: "nat \<Rightarrow> bool" and odd :: "nat \<Rightarrow> bool"
  where
    "even 0 \<longleftrightarrow> True"
  | "even (Suc x) \<longleftrightarrow> odd x"
  | "odd 0 \<longleftrightarrow> False"
  | "odd (Suc x) \<longleftrightarrow> even x"

-- \<open>corresponding mutual induction schema works on two goals simultaneously\<close>
thm even_odd.induct

lemma
  shows "even n \<Longrightarrow> \<exists>k. n = 2 * k"
    and "odd n \<Longrightarrow> \<exists>k. n = 2 * k + 1"
oops

section \<open>A Matching Algorithm\<close>

-- \<open>term map\<close>
type_synonym ('f, 'v) tmap = "'v \<Rightarrow> ('f, 'v) term option"


subsection \<open>Exercises\<close>

text \<open>
  \<bullet> define two mutually recursive functions

    \<open>match_term :: ('f, 'v) tmap \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) tmap option\<close>

    and

    \<open>match_list :: ('f, 'v) tmap \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) tmap option\<close>

    such that \<open>match_term \<sigma> s t\<close> tries to extend the term map \<open>\<sigma>\<close> to
    a matcher between the terms \<open>s\<close> and \<open>t\<close> (and similarly for lists of
    terms with \<open>match_list \<sigma> ss ts\<close>)

  \<bullet> then prove by mutual induction that successful matching yields an
    extension of the given term map to the variables of the pattern term:

    \<open>match_term \<sigma> s t = Some \<tau> \<Longrightarrow> \<sigma> \<subseteq>\<^sub>m \<tau> \<and> dom \<tau> = dom \<sigma> \<union> vars_term s\<close>

    and

    \<open>match_list \<sigma> ss ts = Some \<tau> \<Longrightarrow> \<sigma> \<subseteq>\<^sub>m \<tau> \<and> dom \<tau> = dom \<sigma> \<union> \<Union>(set (map vars_term ss))\<close>
\<close>

fun match_term :: "('f, 'v) tmap \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) tmap option"
and match_list :: "('f, 'v) tmap \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) tmap option"
  where
    "match_term \<sigma> (Var x) t = undefined"
  | "match_term \<sigma> (Fun f ss) (Fun g ts) = undefined"
  | "match_term \<sigma> s t = undefined"
  | "match_list \<sigma> [] [] = undefined"
  | "match_list \<sigma> (s # ss) (t # ts) = undefined"
  | "match_list \<sigma> ss ts = undefined"

lemma match_Some_dom:
  shows "match_term \<sigma> s t = Some \<tau> \<Longrightarrow> \<sigma> \<subseteq>\<^sub>m \<tau> \<and> dom \<tau> = dom \<sigma> \<union> vars_term s"
    and "match_list \<sigma> ss ts = Some \<tau> \<Longrightarrow> \<sigma> \<subseteq>\<^sub>m \<tau> \<and> dom \<tau> = dom \<sigma> \<union> \<Union>(set (map vars_term ss))"
oops

section \<open>Soundness of the Matching Algorithm\<close>

-- \<open>obtaining a substitution from a map\<close>
definition "som m x = (case m x of None \<Rightarrow> Var x | Some t \<Rightarrow> t)"

subsection \<open>Exercises\<close>

text \<open>
  \<bullet> prove that an update of a term map is reflected in the
    corresponding substitution \<open>som (\<sigma>(x \<mapsto> t)) x = t\<close> and
    register this result as simplification rule

  \<bullet> prove \<open>t \<cdot> \<sigma> = t \<cdot> \<tau> \<longleftrightarrow> (\<forall>x \<in> vars_term t. \<sigma> x = \<tau> x)\<close>
    and register this result as simplification rule

  \<bullet> prove that the extension of a term map \<open>\<sigma>\<close> that covers
    already all variables of a term behaves the same as \<open>\<sigma>\<close>:
    \<open>\<sigma> \<subseteq>\<^sub>m \<tau> \<Longrightarrow> vars_term t \<subseteq> dom \<sigma> \<Longrightarrow> t \<cdot> som \<tau> = t \<cdot> som \<sigma>\<close>
\<close>

lemma som_map_upd [simp]: "som (\<sigma>(x \<mapsto> t)) x = t"
  oops

lemma subst_term_eq [simp]:
  "t \<cdot> \<sigma> = t \<cdot> \<tau> \<longleftrightarrow> (\<forall>x \<in> vars_term t. \<sigma> x = \<tau> x)"
  oops

lemma map_le_som_eq:
  "\<sigma> \<subseteq>\<^sub>m \<tau> \<Longrightarrow> vars_term t \<subseteq> dom \<sigma> \<Longrightarrow> t \<cdot> som \<tau> = t \<cdot> som \<sigma>"
  oops


subsection \<open>Exercise\<close>

text \<open>
  \<bullet> show by mutual induction that a successful call to \<open>match_term\<close>
    actually results in a matching substitution

  \<bullet> finally, using the previous results, provide a function

    \<open>match :: ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) subst option\<close>

    that implements matching and show that it is sound
\<close>

end
