theory Term
  imports Main
begin

section \<open>Definition of Types\<close>

subsection \<open>@{command datatype}\<close>

text \<open>
  \<bullet> defines new algebraic datatype as in functional programming

  \<bullet> several automatically generated functions:
    selectors, map, set-functions, etc.

  \<bullet> provides induction scheme, simplification rules, etc.

  \<bullet> @{doc "datatypes"}
\<close>


text \<open>
  Definition of first-order terms for
  function symbols of type @{typ 'f}
  and variables of type @{typ 'v}
\<close>

datatype ('f, vars_term: 'v) "term" =
  Var 'v |
  Fun 'f "('f, 'v) term list"

print_theorems
text \<open>
  Invoke @{command print_theorems} directly
  after a declaration to see new theorem that have been derived
\<close>

text \<open>Induction rule\<close>

thm term.induct

text \<open>Case analysis\<close>

thm term.exhaust

text \<open>
  @{const vars_term} is set-function which collects all
  variables in a term
\<close>

term vars_term


text \<open>Simplification rules\<close>

thm term.simps

subsection \<open>@{command type_synonym}\<close>

text \<open>
  \<bullet> for defining type abbreviations
  \<bullet> just syntactic sugar, no new type is created
  \<bullet> no new theorems will be created
\<close>

type_synonym ('f, 'v) subst = "'v \<Rightarrow> ('f, 'v) term"

print_theorems

subsection \<open>@{command typedef}\<close>

text \<open>
  \<bullet> for defining subtypes
  \<bullet> internally used by @{command datatype}
  \<bullet> not further relevant in this course
\<close>

typedef 'f ground_term = "{t :: ('f, unit) term . vars_term t = {}}"
  by (intro exI [of _ "Fun _ []"]) (auto)


section \<open>Definition of Functions\<close>

text \<open>
  Five different definition principles
  \<bullet> @{command abbreviation}
     \<bullet> similar to type-synonym
     \<bullet> just syntactic sugar

  \<bullet> @{command definition}
     \<bullet> non-recursive
     \<bullet> no pattern matching
     \<bullet> equality not added to simplification rules

  \<bullet> @{command function}, @{command fun}
     \<bullet> recursive
     \<bullet> manual or automatic termination proof
     \<bullet> pattern matching
     \<bullet> induction scheme
     \<bullet> @{doc "functions"}

  \<bullet> @{command primrec}
     \<bullet> primitive recursion
     \<bullet> faster than @{command fun}
     \<bullet> @{doc "datatypes"}

  \<bullet> @{command "partial_function"}
     \<bullet> permits nontermination
     \<bullet> no pattern matching
     \<bullet> syntactic restrictions (tailrec, option)
     \<bullet> @{doc "isar-ref"}, Chapter 11.2.2
  \<close>

primrec subst_term :: "('f, 'v) subst \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
  where
    "subst_term \<sigma> (Var x) = \<sigma> x"
  | "subst_term \<sigma> (Fun f ss) = Fun f (map (subst_term \<sigma>) ss)"

print_theorems

abbreviation apply_subst_term (infixl "\<cdot>" 67)
  where
    "t \<cdot> \<sigma> \<equiv> subst_term \<sigma> t"

thm subst_term.simps

section \<open>Exercises\<close>

text \<open>
  \<bullet> define a datatype "('f,'v) ctxt" for single-hole contexts

  \<bullet> define a function @{term ctxt_term} which takes
    a context and a term, and inserts the term into the hole.
    \<bullet> study @{file "~~/src/HOL/List.thy"} to find auxiliary functions on lists
    \<bullet> use @{command find_consts} to find auxiliary functions on lists,
      it takes a type as argument with _ as wildcard,
      cf. {@doc "isar-ref"}, Chapter 3.4

  \<bullet> define a function @{term ctxt_compose} which composes two contexts

  \<bullet> define a term which defines term rewriting
    \<bullet> create a type for rewrite rules, i.e., pairs of term, cf. @{typ "'a \<times> 'b"}
    \<bullet> create a type for TRSs, i.e., sets of rewrite rules, cf. @{typ "'a set"}"
    \<bullet> define a constant @{term rewrite_relation} which takes as input some TRS R
        and returns some object of type TRS, namely the rewrite relation of R.
        \<bullet> use set-comprehension like @{term "{ (f a, g b) | a b. a \<in> X \<and> p a b }"}
\<close>

datatype ('f, 'v) ctxt =
  Hole ("\<box>") |
  More 'f "('f, 'v) term list" "('f, 'v) ctxt" "('f, 'v) term list"

text \<open>To search for append\<close>
find_consts "?'a list \<Rightarrow> ?'a list \<Rightarrow> ?'a list"

text \<open>To search for Cons, we just ctrl-click on @{typ "'a list"}\<close>

primrec ctxt_term :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term" ("_\<langle>_\<rangle>" [1000, 0] 1000)
  where
    "\<box>\<langle>s\<rangle> = s"
  | "(More f ss\<^sub>1 C ss\<^sub>2)\<langle>s\<rangle> = Fun f (ss\<^sub>1 @ C\<langle>s\<rangle> # ss\<^sub>2)"

text \<open>
  Syntax annotations like the ones for @{const Hole} and @{const ctxt_term}
  are described in more detail in @{doc "isar-ref"}, Chapter 8.2
\<close>

fun
  ctxt_compose :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt" (infixl "\<circ>\<^sub>c" 75)
  where
    "\<box> \<circ>\<^sub>c D = D"
  | "(More f ss1 C ss2) \<circ>\<^sub>c D = More f ss1 (C \<circ>\<^sub>c D) ss2"

type_synonym ('f, 'v) rule = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) trs = "('f, 'v) rule set"

definition rewrite_relation :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs"
  where
    "rewrite_relation R = { (C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) | l r C \<sigma>. (l, r) \<in> R }"

print_theorems

text \<open>Alternative definition\<close>
lemma "rewrite_relation R =
  { (s, t) . \<exists> l r C \<sigma>. s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<and> (l, r) \<in> R }"
  unfolding rewrite_relation_def by auto

end
